# Generated by Django 4.1.8 on 2023-05-04 02:42

from django.db import migrations, models
import taggit.managers
import utilities.json


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('extras', '0092_delete_jobresult'),
    ]

    operations = [
        migrations.CreateModel(
            name='Server',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('name', models.CharField(max_length=255, unique=True)),
                ('server_url', models.CharField(max_length=255)),
                ('username', models.CharField(blank=True, max_length=255, null=True)),
                ('password', models.CharField(blank=True, max_length=255, null=True)),
                ('ssl_verify', models.BooleanField(default=True)),
                ('client_cert_path', models.CharField(blank=True, max_length=4096, null=True)),
                ('client_key_path', models.CharField(blank=True, max_length=4096, null=True)),
                ('ca_file_path', models.CharField(blank=True, max_length=4096, null=True)),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from collections.abc import Sequence
from typing import Any, TypedDict

import requests
from requests.models import HTTPBasicAuth


class KeaResponse(TypedDict):
    result: int
    arguments: dict[str, Any] | None
    text: str | None


class KeaClient:
    def __init__(
        self,
        url: str,
        username: str | None = None,
        password: str | None = None,
        verify: bool | str | None = None,
        client_cert: str | None = None,
        client_key: str | None = None,
        timeout: int = 30,
    ):
        if (client_cert is not None and client_key is None) or (
            client_cert is None and client_key is not None
        ):
            raise ValueError("Key and Cert must be used together.")

        self.url = url
        self.timeout = timeout

        self._session = requests.Session()
        if verify is not None:
            self._session.verify = verify
        if username is not None and password is not None:
            self._session.auth = HTTPBasicAuth(username, password)
        if client_cert is not None and client_key is not None:
            self._session.cert = (client_cert, client_key)

    def command(
        self,
        command: str,
        service: list[str] | None = None,
        arguments: dict[str, Any] | None = None,
        check: None | Sequence[int] = (0,),
    ) -> list[KeaResponse]:
        body: dict[str, Any] = {"command": command}

        if service is not None:
            body["service"] = service

        if arguments is not None:
            body["arguments"] = arguments

        resp = self._session.post(self.url, json=body, timeout=self.timeout)
        resp.raise_for_status()
        resp_json = resp.json()
        assert isinstance(resp_json, list)
        if check is not None:
            check_response(resp_json, check)
        return resp_json


class KeaException(Exception):
    def __init__(
        self, resp: KeaResponse, msg: str | None = None, index: int | None = None
    ) -> None:
        self.index = index
        self.response = resp

        if msg is None:
            msg = f"Kea returned result[{index}] {self.response.get('result')}"
        message = f"{msg}: {self.response.get('text')}"
        super().__init__(message)


def check_response(resp: list[KeaResponse], ok_codes: Sequence[int]) -> None:
    """Raise a KeaException for any non 0 responses."""
    for idx, kr in enumerate(resp):
        if kr["result"] not in ok_codes:
            raise KeaException(kr, index=idx)



BY_IP = "ip"
BY_HOSTNAME = "hostname"
BY_DUID = "duid"
BY_SUBNET = "subnet"
BY_SUBNET_ID = "subnet_id"
BY_HW_ADDRESS = "hw"
BY_CLIENT_ID = "client_id"

HEX_STRING_REGEX = r"^([0-9A-Fa-f]{2}[:-]?)*([0-9A-Fa-f]{2})$"

# kea/src/lib/dhcp
# RFC8415 section 11.1
DUID_MAX_OCTETS = 128
DUID_MIN_OCTETS = 1
CLIENT_ID_MAX_OCTETS = DUID_MAX_OCTETS
CLIENT_ID_MIN_OCTETS = 2



import logging
from abc import ABCMeta
from typing import Any, Generic, TypeVar

from django.contrib import messages
from django.http import HttpResponse, HttpResponseForbidden
from django.http.request import HttpRequest
from django.shortcuts import redirect, render
from django.urls import reverse
from netaddr import IPAddress, IPNetwork
from netbox.tables import BaseTable
from netbox.views import generic
from utilities.exceptions import AbortRequest
from utilities.htmx import htmx_partial
from utilities.paginator import EnhancedPaginator, get_paginate_count
from utilities.views import GetReturnURLMixin, ViewTab, register_model_view

from . import constants, forms, tables
from .filtersets import ServerFilterSet
from .kea import KeaClient
from .models import Server
from .utilities import (
    OptionalViewTab,
    check_dhcp_enabled,
    export_table,
    format_duration,
    format_leases,
)

T = TypeVar("T", bound=BaseTable)


@register_model_view(Server)
class ServerView(generic.ObjectView):
    queryset = Server.objects.all()


@register_model_view(Server, "edit")
class ServerEditView(generic.ObjectEditView):
    queryset = Server.objects.all()
    form = forms.ServerForm


@register_model_view(Server, "delete")
class ServerDeleteView(generic.ObjectDeleteView):
    queryset = Server.objects.all()


class ServerListView(generic.ObjectListView):
    queryset = Server.objects.all()
    table = tables.ServerTable
    filterset = ServerFilterSet
    filterset_form = forms.ServerFilterForm


class ServerBulkDeleteView(generic.BulkDeleteView):
    queryset = Server.objects.all()
    table = tables.ServerTable


@register_model_view(Server, "status")
class ServerStatusView(generic.ObjectView):
    queryset = Server.objects.all()
    tab = ViewTab(label="Status", weight=1000)
    template_name = "netbox_kea/server_status.html"

    def _get_ca_status(self, client: KeaClient) -> dict[str, Any]:
        """Get the control agent status"""
        status = client.command("status-get")
        args = status[0]["arguments"]
        assert args is not None

        version = client.command("version-get")
        version_args = version[0]["arguments"]
        assert version_args is not None

        return {
            "PID": args["pid"],
            "Uptime": format_duration(int(args["uptime"])),
            "Time since reload": format_duration(int(args["reload"])),
            "Version": version_args["extended"],
        }

    def _get_dhcp_status(
        self, server: Server, client: KeaClient
    ) -> dict[str, dict[str, Any]]:
        resp: dict[str, dict[str, Any]] = {}

        # Map of name to pretty name
        service_names = {"dhcp6": "DHCPv6", "dhcp4": "DHCPv4"}
        services = []
        if server.dhcp6:
            services.append("dhcp6")
        if server.dhcp4:
            services.append("dhcp4")
        service_keys = list(services)

        dhcp_status = client.command("status-get", service=service_keys)
        dhcp_version = client.command("version-get", service=service_keys)
        assert len(dhcp_status) == len(services)
        assert len(dhcp_version) == len(services)
        for svc, status, version in zip(services, dhcp_status, dhcp_version):
            args = status["arguments"]
            assert args is not None

            version_args = version["arguments"]
            assert version_args is not None

            resp[service_names[svc]] = {
                "PID": args["pid"],
                "Uptime": format_duration(args["uptime"]),
                "Time since reload": format_duration(int(args["reload"])),
                "Version": version_args["extended"],
            }

            if (ha := args.get("high-availability")) is not None:
                # https://kea.readthedocs.io/en/latest/arm/hooks.html#load-balancing-configuration
                # Note that while the top-level parameter high-availability is a list,
                # only a single entry is currently supported.

                ha_servers = ha[0].get("ha-servers")
                ha_local = ha_servers.get("local", {})
                ha_remote = ha_servers.get("remote", {})
                resp[service_names[svc]].update(
                    {
                        "HA mode": ha[0].get("ha-mode"),
                        "HA local role": ha_local.get("role"),
                        "HA local state": ha_local.get("state"),
                        "HA remote connection interrupted": str(
                            ha_remote.get("connection-interrupted")
                        ),
                        "HA remote age (seconds)": ha_remote.get("age"),
                        "HA remote role": ha_remote.get("role"),
                        "HA remote last state": ha_remote.get("last-state"),
                        "HA remote in touch": ha_remote.get("in-touch"),
                        "HA remote unacked clients": ha_remote.get("unacked-clients"),
                        "HA remote unacked clients left": ha_remote.get(
                            "unacked-clients-left"
                        ),
                        "HA remote connecting clients": ha_remote.get(
                            "connecting-clients"
                        ),
                    }
                )
        return resp

    def _get_statuses(
        self, server: Server, client: KeaClient
    ) -> dict[str, dict[str, Any]]:
        return {
            "Control Agent": self._get_ca_status(client),
            **self._get_dhcp_status(server, client),
        }

    def get_extra_context(
        self, request: HttpResponse, instance: Server
    ) -> dict[str, Any]:
        return {"statuses": self._get_statuses(instance, instance.get_client())}


class BaseServerLeasesView(generic.ObjectView, Generic[T]):
    template_name = "netbox_kea/server_dhcp_leases.html"
    queryset = Server.objects.all()
    table: type[T]

    def get_table(self, data: list[dict[str, Any]], request: HttpRequest) -> T:
        table = self.table(data, user=request.user)
        table.configure(request)
        return table

    def get_leases_page(
        self, client: KeaClient, subnet: IPNetwork, page: str | None, per_page: int
    ) -> tuple[list[dict[str, Any]], str | None]:
        if page:
            frm = page
        elif int(subnet.network) == 0:
            frm = str(subnet.network)
        else:
            frm = str(subnet.network - 1)

        resp = client.command(
            f"lease{self.dhcp_version}-get-page",
            service=[f"dhcp{self.dhcp_version}"],
            arguments={"from": frm, "limit": per_page},
            check=(0, 3),
        )

        if resp[0]["result"] == 3:
            return [], None

        args = resp[0]["arguments"]
        assert args is not None

        raw_leases = args["leases"]
        next = f"{raw_leases[-1]['ip-address']}" if args["count"] == per_page else None
        for i, lease in enumerate(raw_leases):
            lease_ip = IPAddress(lease["ip-address"])
            if lease_ip not in subnet:
                raw_leases = raw_leases[:i]
                next = None
                break

        subnet_leases = format_leases(raw_leases)

        return subnet_leases, next

    def get_leases(self, client: KeaClient, q: Any, by: str) -> list[dict[str, Any]]:
        arguments: dict[str, Any]
        command = ""
        multiple = True

        if by == constants.BY_IP:
            arguments = {"ip-address": q}
            multiple = False
        elif by == constants.BY_HW_ADDRESS:
            arguments = {"hw-address": q}
            command = "-by-hw-address"
        elif by == constants.BY_HOSTNAME:
            arguments = {"hostname": q}
            command = "-by-hostname"
        elif by == constants.BY_CLIENT_ID:
            arguments = {"client-id": q}
            command = "-by-client-id"
        elif by == constants.BY_SUBNET_ID:
            command = "-all"
            arguments = {"subnets": [int(q)]}
        elif by == constants.BY_DUID:
            command = "-by-duid"
            arguments = {"duid": q}
        else:
            # We should never get here because the
            # form should of been validated.
            raise AbortRequest(f"Invalid search by (this shouldn't happen): {by}")
        resp = client.command(
            f"lease{self.dhcp_version}-get{command}",
            service=[f"dhcp{self.dhcp_version}"],
            arguments=arguments,
            check=(0, 3),
        )

        if resp[0]["result"] == 3:
            return []

        args = resp[0]["arguments"]
        assert args is not None
        if multiple is True:
            return format_leases(args["leases"])
        return format_leases([args])

    def get_extra_context(
        self, request: HttpRequest, _instance: Server
    ) -> dict[str, Any]:
        # For non-htmx requests.

        table = self.get_table([], request)
        form = self.form(request.GET) if "q" in request.GET else self.form()
        return {"form": form, "table": table}

    def get_export(self, request: HttpRequest, **kwargs) -> HttpResponse:
        form = self.form(request.GET)
        if not form.is_valid():
            messages.warning(request, "Invalid form for export.")
            return redirect(request.path)

        instance = self.get_object(**kwargs)

        by = form.cleaned_data["by"]
        q = form.cleaned_data["q"]
        client = instance.get_client()
        if by == constants.BY_SUBNET:
            leases = []
            page: str | None = ""  # start from the beginning
            while page is not None:
                page_leases, page = self.get_leases_page(
                    client,
                    q,
                    page,
                    per_page=get_paginate_count(request),
                )
                leases += page_leases
        else:
            leases = self.get_leases(client, q, by)

        table = self.get_table(leases, request)
        return export_table(
            table, "leases.csv", use_selected_columns=request.GET["export"] == "table"
        )

    def get(self, request: HttpRequest, **kwargs) -> HttpResponse:
        logger = logging.getLogger("netbox_kea.views.BaseServerDHCPLeasesView")

        instance: Server = self.get_object(**kwargs)

        if resp := check_dhcp_enabled(instance, self.dhcp_version):
            return resp

        if "export" in request.GET:
            return self.get_export(request, **kwargs)

        if not request.htmx:
            return super().get(request, **kwargs)

        try:
            form = self.form(request.GET)
            if not form.is_valid():
                table = self.get_table([], request)
                return render(
                    request,
                    "netbox_kea/server_dhcp_leases_htmx.html",
                    {
                        "is_embedded": False,
                        "form": form,
                        "table": table,
                        "paginate": False,
                    },
                )

            by = form.cleaned_data["by"]
            q = form.cleaned_data["q"]
            client = instance.get_client()
            if by == "subnet":
                leases, next_page = self.get_leases_page(
                    client,
                    q,
                    form.cleaned_data["page"],
                    per_page=get_paginate_count(request),
                )
                paginate = True
            else:
                paginate = False
                next_page = None
                leases = self.get_leases(client, q, by)

            table = self.get_table(leases, request)

            can_delete = request.user.has_perm(
                "netbox_kea.bulk_delete_lease_from_server",
                obj=instance,
            )
            if not can_delete:
                table.columns.hide("pk")

            return render(
                request,
                "netbox_kea/server_dhcp_leases_htmx.html",
                {
                    "can_delete": can_delete,
                    "is_embedded": False,
                    "delete_action": reverse(
                        f"plugins:netbox_kea:server_leases{self.dhcp_version}_delete",
                        args=[instance.pk],
                    ),
                    "form": form,
                    "table": table,
                    "next_page": next_page,
                    "paginate": paginate,
                    "page_lengths": EnhancedPaginator.default_page_lengths,
                },
            )
        except Exception as e:
            logger.exception("exception on DHCP leases HTMX handler")
            return render(
                request,
                "netbox_kea/exception_htmx.html",
                {"type_": type(e).__name__, "exception": str(e)},
            )


@register_model_view(Server, "leases6")
class ServerLeases6View(BaseServerLeasesView[tables.LeaseTable6]):
    tab = OptionalViewTab(
        label="DHCPv6 Leases", weight=1010, is_enabled=lambda s: s.dhcp6
    )
    form = forms.Leases6SearchForm
    table = tables.LeaseTable6
    dhcp_version = 6


@register_model_view(Server, "leases4")
class ServerLeases4View(BaseServerLeasesView[tables.LeaseTable4]):
    tab = OptionalViewTab(
        label="DHCPv4 Leases", weight=1020, is_enabled=lambda s: s.dhcp4
    )
    form = forms.Leases4SearchForm
    table = tables.LeaseTable4
    dhcp_version = 4


class FakeLeaseModelMeta:
    verbose_name_plural = "leases"


# Fake model to allow us to use the bulk_delete.html template.
class FakeLeaseModel:
    _meta = FakeLeaseModelMeta


class BaseServerLeasesDeleteView(
    GetReturnURLMixin, generic.ObjectView, metaclass=ABCMeta
):
    queryset = Server.objects.all()
    default_return_url = "plugins:netbox_kea:server_list"

    def delete_lease(self, client: KeaClient, ip: str) -> None:
        client.command(
            f"lease{self.dhcp_version}-del",
            arguments={"ip-address": ip},
            service=[f"dhcp{self.dhcp_version}"],
            check=(0, 3),
        )

    def get(self, request: HttpRequest, **kwargs):
        return redirect(self.get_return_url(request, obj=self.get_object(**kwargs)))

    def post(self, request: HttpRequest, **kwargs) -> HttpResponse:
        instance: Server = self.get_object(**kwargs)

        if not request.user.has_perm(
            "netbox_kea.bulk_delete_lease_from_server", obj=instance
        ):
            return HttpResponseForbidden(
                "This user does not have permission to delete DHCP leases."
            )

        form = self.form(request.POST)

        if not form.is_valid():
            messages.warning(request, str(form.errors))
            return redirect(self.get_return_url(request, obj=instance))

        lease_ips = form.cleaned_data["pk"]
        if "_confirm" not in request.POST:
            return render(
                request,
                "generic/bulk_delete.html",
                {
                    "model": FakeLeaseModel,
                    "table": tables.LeaseDeleteTable(
                        ({"ip": ip} for ip in lease_ips),
                        orderable=False,
                    ),
                    "form": form,
                    "return_url": self.get_return_url(request, obj=instance),
                },
            )

        client = instance.get_client()

        for ip in lease_ips:
            try:
                self.delete_lease(client, ip)
            except Exception as e:  # noqa: PERF203
                messages.error(request, f"Error deleting lease {ip}: {repr(e)}")
                return redirect(self.get_return_url(request, obj=instance))

        messages.success(
            request, f"Deleted {len(lease_ips)} DHCPv{self.dhcp_version} lease(s)."
        )
        return redirect(self.get_return_url(request, obj=instance))


class ServerLeases6DeleteView(BaseServerLeasesDeleteView):
    form = forms.Lease6DeleteForm
    dhcp_version = 6


class ServerLeases4DeleteView(BaseServerLeasesDeleteView):
    form = forms.Lease4DeleteForm
    dhcp_version = 4


class BaseServerDHCPSubnetsView(generic.ObjectChildrenView):
    table = tables.SubnetTable
    queryset = Server.objects.all()
    template_name = "netbox_kea/server_dhcp_subnets.html"

    def get_children(
        self, request: HttpRequest, parent: Server
    ) -> list[dict[str, Any]]:
        return self.get_subnets(parent)

    def get_subnets(self, server: Server) -> list[dict[str, Any]]:
        client = server.get_client()
        config = client.command("config-get", service=[f"dhcp{self.dhcp_version}"])
        assert config[0]["arguments"] is not None
        subnets = config[0]["arguments"][f"Dhcp{self.dhcp_version}"][
            f"subnet{self.dhcp_version}"
        ]
        subnet_list = [
            {
                "id": s["id"],
                "subnet": s["subnet"],
                "dhcp_version": self.dhcp_version,
                "server_pk": server.pk,
            }
            for s in subnets
            if "id" in s and "subnet" in s
        ]

        for sn in config[0]["arguments"][f"Dhcp{self.dhcp_version}"]["shared-networks"]:
            subnet_list.extend(
                {
                    "id": s["id"],
                    "subnet": s["subnet"],
                    "shared_network": sn["name"],
                    "dhcp_version": self.dhcp_version,
                    "server_pk": server.pk,
                }
                for s in sn[f"subnet{self.dhcp_version}"]
            )

        return subnet_list

    def get(self, request: HttpRequest, **kwargs: Any) -> HttpResponse:
        instance = self.get_object(**kwargs)
        if resp := check_dhcp_enabled(instance, self.dhcp_version):
            return resp

        # We can't use the original get() since it calls get_table_configs which requires a NetBox model.
        instance = self.get_object(**kwargs)
        child_objects = self.get_children(request, instance)

        table_data = self.prep_table_data(request, child_objects, instance)
        table = self.get_table(table_data, request, False)

        if "export" in request.GET:
            return export_table(
                table,
                filename=f"kea-dhcpv{self.dhcp_version}-subnets.csv",
                use_selected_columns=request.GET["export"] == "table",
            )

        # If this is an HTMX request, return only the rendered table HTML
        if htmx_partial(request):
            return render(
                request,
                "htmx/table.html",
                {
                    "object": instance,
                    "table": table,
                    "model": self.child_model,
                },
            )

        return render(
            request,
            self.get_template_name(),
            {
                "object": instance,
                "base_template": f"{instance._meta.app_label}/{instance._meta.model_name}.html",
                "table": table,
                "table_config": f"{table.name}_config",
                "return_url": request.get_full_path(),
            },
        )


@register_model_view(Server, "subnets6")
class ServerDHCP6SubnetsView(BaseServerDHCPSubnetsView):
    tab = OptionalViewTab(
        label="DHCPv6 Subnets", weight=1030, is_enabled=lambda s: s.dhcp6
    )
    dhcp_version = 6


@register_model_view(Server, "subnets4")
class ServerDHCP4SubnetsView(BaseServerDHCPSubnetsView):
    tab = OptionalViewTab(
        label="DHCPv4 Subnets", weight=1040, is_enabled=lambda s: s.dhcp4
    )
    dhcp_version = 4



from netbox.api.serializers import NetBoxModelSerializer
from rest_framework import serializers

from ..models import Server


class ServerSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="plugins-api:netbox_kea-api:server-detail"
    )

    class Meta:
        model = Server
        fields = (
            "id",
            "name",
            "server_url",
            "username",
            "password",
            "ssl_verify",
            "client_cert_path",
            "client_key_path",
            "ca_file_path",
            "dhcp6",
            "dhcp4",
            "url",
            "display",
            "tags",
            "last_updated",
        )
        brief_fields = ("id", "url", "name", "server_url")



# Generated by Django 4.1.8 on 2023-05-15 04:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_kea', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='server',
            options={'ordering': ('name',)},
        ),
        migrations.AddField(
            model_name='server',
            name='dhcp4',
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name='server',
            name='dhcp6',
            field=models.BooleanField(default=True),
        ),
    ]



from netbox.api.viewsets import NetBoxModelViewSet

from .. import filtersets, models
from .serializers import ServerSerializer


class ServerViewSet(NetBoxModelViewSet):
    queryset = models.Server.objects.prefetch_related("tags")
    filterset_class = filtersets.ServerFilterSet
    serializer_class = ServerSerializer



from netbox.filtersets import NetBoxModelFilterSet

from .models import Server


class ServerFilterSet(NetBoxModelFilterSet):
    class Meta:
        model = Server
        fields = ("id", "name", "server_url", "dhcp4", "dhcp6")



from netbox.plugins import PluginMenuButton, PluginMenuItem

menu_items = (
    PluginMenuItem(
        link="plugins:netbox_kea:server_list",
        link_text="Servers",
        permissions=["netbox_kea.view_server"],
        buttons=(
            PluginMenuButton(
                link="plugins:netbox_kea:server_add",
                title="Add",
                icon_class="mdi mdi-plus-thick",
                permissions=["netbox_kea.add_server"],
            ),
        ),
    ),
)



from typing import Any, Literal

from django import forms
from django.core.exceptions import ValidationError
from netaddr import EUI, AddrFormatError, IPAddress, IPNetwork, mac_unix_expanded
from netbox.forms import NetBoxModelFilterSetForm, NetBoxModelForm
from utilities.forms import BOOLEAN_WITH_BLANK_CHOICES
from utilities.forms.fields import TagFilterField

from . import constants
from .models import Server
from .utilities import is_hex_string


class ServerForm(NetBoxModelForm):
    class Meta:
        model = Server
        fields = (
            "name",
            "server_url",
            "username",
            "password",
            "ssl_verify",
            "client_cert_path",
            "client_key_path",
            "ca_file_path",
            "dhcp6",
            "dhcp4",
            "tags",
        )


class VeryHiddenInput(forms.HiddenInput):
    """Returns an empty string on render."""

    input_type = "hidden"
    template_name = ""

    def render(self, name: str, value: Any, attrs: Any, renderer: Any) -> str:
        return ""


class ServerFilterForm(NetBoxModelFilterSetForm):
    model = Server
    tag = TagFilterField(model)
    dhcp4 = forms.NullBooleanField(
        label="DHCPv4",
        required=False,
        widget=forms.Select(choices=BOOLEAN_WITH_BLANK_CHOICES),
    )
    dhcp6 = forms.NullBooleanField(
        label="DHCPv6",
        required=False,
        widget=forms.Select(choices=BOOLEAN_WITH_BLANK_CHOICES),
    )


class BaseLeasesSarchForm(forms.Form):
    q = forms.CharField(label="Search")
    page = forms.CharField(required=False, widget=VeryHiddenInput)

    def clean(self) -> dict[str, Any] | None:
        ip_version = self.Meta.ip_version
        cleaned_data = super().clean()
        q = cleaned_data.get("q")
        by = cleaned_data.get("by")

        if q and not by:
            raise ValidationError({"by": "Search attribute is empty."})
        elif by and not q:
            raise ValidationError({"q": "Search value is empty."})

        if by == constants.BY_SUBNET:
            try:
                if "/" not in q:
                    raise ValidationError({"q": "CIDR mask is required"})
                net = IPNetwork(q, version=ip_version)
                if net.ip != net.cidr.ip:
                    raise ValidationError(
                        {"q": f"{net} is not a valid prefix. Did you mean {net.cidr}?"}
                    )
                cleaned_data["q"] = net
            except (AddrFormatError, TypeError, ValueError) as e:
                raise ValidationError({"q": f"Invalid IPv{ip_version} subnet."}) from e
        elif by == constants.BY_SUBNET_ID:
            try:
                i = int(q)
                if i <= 0:
                    raise ValidationError({"q": "Invalid subnet ID."})
                cleaned_data["q"] = i
            except ValueError as e:
                raise ValidationError({"q": "Subnet ID must be an integer."}) from e
        elif by == constants.BY_IP:
            try:
                # use IPAddress to normalize values
                cleaned_data["q"] = str(IPAddress(q, version=ip_version))
            except (AddrFormatError, TypeError, ValueError) as e:
                raise ValidationError({"q": f"Invalid IPv{ip_version} address."}) from e
        elif by == constants.BY_HW_ADDRESS:
            try:
                cleaned_data["q"] = str(EUI(q, version=48, dialect=mac_unix_expanded))
            except (AddrFormatError, TypeError, ValueError) as e:
                raise ValidationError({"q": "Invalid hardware address."}) from e
        elif by == constants.BY_DUID:
            if not is_hex_string(
                q, constants.DUID_MIN_OCTETS, constants.DUID_MAX_OCTETS
            ):
                raise ValidationError({"q": "Invalid DUID."})
            cleaned_data["q"] = q.replace("-", "")
        elif by == constants.BY_CLIENT_ID:
            if not is_hex_string(
                q, constants.CLIENT_ID_MIN_OCTETS, constants.DUID_MAX_OCTETS
            ):
                raise ValidationError({"q": "Invalid client ID."})
            cleaned_data["q"] = q.replace("-", "")

        page = cleaned_data["page"]
        if page:
            if by != constants.BY_SUBNET:
                raise ValidationError({"page": "page is only supported with subnet."})
            try:
                page_ip = IPAddress(page, version=ip_version)
                if page_ip not in cleaned_data["q"]:
                    raise ValidationError({"page": "page is not in the given subnet"})

                cleaned_data["page"] = str(page_ip)
            except AddrFormatError as e:
                raise ValidationError({"page": "Invalid IP."}) from e


class Leases4SearchForm(BaseLeasesSarchForm):
    by = forms.ChoiceField(
        label="Attribute",
        choices=(
            (constants.BY_IP, "IP Address"),
            (constants.BY_HOSTNAME, "Hostname"),
            (constants.BY_HW_ADDRESS, "Hardware Address"),
            (constants.BY_CLIENT_ID, "Client ID"),
            (constants.BY_SUBNET, "Subnet"),
            (constants.BY_SUBNET_ID, "Subnet ID"),
        ),
        required=True,
    )

    class Meta:
        ip_version = 4


class Leases6SearchForm(BaseLeasesSarchForm):
    by = forms.ChoiceField(
        label="Attribute",
        choices=(
            (constants.BY_IP, "IP Address"),
            (constants.BY_HOSTNAME, "Hostname"),
            (constants.BY_DUID, "DUID"),
            (constants.BY_SUBNET, "Subnet"),
            (constants.BY_SUBNET_ID, "Subnet ID"),
        ),
        required=True,
    )

    class Meta:
        ip_version = 6


class MultipleIPField(forms.MultipleChoiceField):
    def __init__(self, version: Literal[6, 4], *args, **kwargs) -> None:
        self._version = version
        super().__init__(*args, widget=forms.MultipleHiddenInput, **kwargs)

    def clean(self, value: Any) -> Any:
        if not isinstance(value, list):
            raise forms.ValidationError(f"Expected a list, got {type(value)}.")

        if len(value) == 0:
            raise forms.ValidationError("IP address list is empty.")

        try:
            return [str(IPAddress(ip, version=self._version)) for ip in value]
        except (AddrFormatError, ValueError) as e:
            raise forms.ValidationError("Invalid IP address.") from e


class BaseLeaseDeleteForm(forms.Form):
    # NetBox v4.4 requires a background_job field for the bulk_delete.html
    # template.
    background_job = forms.CharField(
        required=False, widget=VeryHiddenInput, label="background_job"
    )
    return_url = forms.CharField(
        required=False,
        widget=forms.HiddenInput(),
    )


class Lease6DeleteForm(BaseLeaseDeleteForm):
    pk = MultipleIPField(6)


class Lease4DeleteForm(BaseLeaseDeleteForm):
    pk = MultipleIPField(4)



from django.urls import include, path
from utilities.urls import get_model_urls

from . import views

urlpatterns = (
    path("servers/", views.ServerListView.as_view(), name="server_list"),
    path("servers/add/", views.ServerEditView.as_view(), name="server_add"),
    path(
        "servers/delete/",
        views.ServerBulkDeleteView.as_view(),
        name="server_bulk_delete",
    ),
    path(
        "servers/<int:pk>/leases6/delete/",
        views.ServerLeases6DeleteView.as_view(),
        name="server_leases6_delete",
    ),
    path(
        "servers/<int:pk>/leases4/delete/",
        views.ServerLeases4DeleteView.as_view(),
        name="server_leases4_delete",
    ),
    path("servers/<int:pk>/", include(get_model_urls("netbox_kea", "server"))),
)



import strawberry
import strawberry_django
from netbox.graphql.types import NetBoxObjectType

from . import models


@strawberry_django.type(
    models.Server,
    fields="__all__",
)
class ServerType(NetBoxObjectType):
    pass


@strawberry.type
class Query:
    @strawberry.field
    def server(self, id: int) -> ServerType:
        return models.Server.objects.get(pk=id)

    server_list: list[ServerType] = strawberry_django.field()


schema = [Query]



from netbox.plugins import PluginConfig


class NetBoxKeaConfig(PluginConfig):
    name = "netbox_kea"
    verbose_name = "Kea"
    description = "Kea integration for NetBox"
    version = "1.0.3"
    base_url = "kea"
    default_settings = {"kea_timeout": 30}


config = NetBoxKeaConfig



import os

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from netbox.models import NetBoxModel

from .kea import KeaClient


class Server(NetBoxModel):
    name = models.CharField(unique=True, max_length=255)
    server_url = models.CharField(verbose_name="Server URL", max_length=255)
    username = models.CharField(null=True, blank=True, max_length=255)
    password = models.CharField(null=True, blank=True, max_length=255)
    ssl_verify = models.BooleanField(
        default=True,
        verbose_name="SSL Verification",
        help_text="Enable SSL certificate verification. Disable with caution!",
    )
    client_cert_path = models.CharField(
        max_length=4096,
        null=True,
        blank=True,
        verbose_name="Client Certificate",
        help_text="Optional client certificate.",
    )
    client_key_path = models.CharField(
        max_length=4096,
        null=True,
        blank=True,
        verbose_name="Private Key",
        help_text="Optional client key.",
    )
    ca_file_path = models.CharField(
        max_length=4096,
        null=True,
        blank=True,
        verbose_name="CA File Path",
        help_text="The specific CA certificate file to use for SSL verification.",
    )
    dhcp6 = models.BooleanField(verbose_name="DHCPv6", default=True)
    dhcp4 = models.BooleanField(verbose_name="DHCPv4", default=True)

    class Meta:
        ordering = ("name",)

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse("plugins:netbox_kea:server", args=[self.pk])

    def get_client(self) -> KeaClient:
        return KeaClient(
            url=self.server_url,
            username=self.username,
            password=self.password,
            verify=self.ca_file_path or self.ssl_verify,
            client_cert=self.client_cert_path or None,
            client_key=self.client_key_path or None,
            timeout=settings.PLUGINS_CONFIG["netbox_kea"]["kea_timeout"],
        )

    def clean(self) -> None:
        super().clean()

        if self.dhcp4 is False and self.dhcp6 is False:
            raise ValidationError(
                {"dhcp6": "At one of DHCPv4 and DHCPv6 needs to be enabled."}
            )

        if (self.client_cert_path and not self.client_key_path) or (
            not self.client_cert_path and self.client_key_path
        ):
            raise ValidationError(
                {
                    "client_cert_path": "Client certificate and client private key must be used together."
                }
            )

        if self.client_cert_path and not os.path.isfile(self.client_cert_path):
            raise ValidationError(
                {"client_cert_path": "Client certificate doesn't exist."}
            )
        if self.client_key_path and not os.path.isfile(self.client_key_path):
            raise ValidationError(
                {"client_key_path": "Client private key doesn't exist."}
            )

        if self.ca_file_path and not self.ssl_verify:
            raise ValidationError(
                {
                    "ca_file_path": "Cannot specify a CA file when SSL verification is disabled."
                }
            )

        client = self.get_client()
        if self.dhcp6:
            try:
                client.command("version-get", service=["dhcp6"])
            except Exception as e:
                raise ValidationError(
                    {"dhcp6": f"Unable to get DHCPv6 version: {repr(e)}"}
                ) from e
        if self.dhcp4:
            try:
                client.command("version-get", service=["dhcp4"])
            except Exception as e:
                raise ValidationError(
                    {"dhcp4": f"Unable to get DHCPv4 version: {repr(e)}"}
                ) from e



import re
from collections.abc import Callable
from datetime import datetime
from typing import Any, Literal

from django.http import HttpResponse
from django.shortcuts import redirect
from django_tables2 import Table
from django_tables2.export import TableExport
from utilities.views import ViewTab

from . import constants
from .models import Server


def format_duration(s: int | None) -> str | None:
    if s is None:
        return None
    hours, rest = divmod(s, 3600)
    minutes, seconds = divmod(rest, 60)
    return f"{hours:02}:{minutes:02}:{seconds:02}"


def _enrich_lease(now: datetime, lease: dict[str, Any]) -> dict[str, Any]:
    """Add expires at and expires in to a lease."""

    # Need to replace "-" so we can access the values in a template
    lease = {k.replace("-", "_"): v for k, v in lease.items()}
    if "cltt" not in lease and "valid_lft" not in lease:
        return lease

    # https://kea.readthedocs.io/en/kea-2.2.0/arm/hooks.html?highlight=cltt#the-lease4-get-lease6-get-commands
    cltt = lease["cltt"]
    valid_lft = lease["valid_lft"]
    assert isinstance(cltt, int)
    assert isinstance(valid_lft, int)
    expires_at = datetime.fromtimestamp(cltt + valid_lft)
    lease["expires_at"] = expires_at
    lease["expires_in"] = (expires_at - now).seconds
    lease["cltt"] = datetime.fromtimestamp(cltt)
    return lease


def format_leases(leases: list[dict[str, Any]]) -> list[dict[str, Any]]:
    now = datetime.now()
    return [_enrich_lease(now, ls) for ls in leases]


def export_table(
    table: Table,
    filename: str,
    use_selected_columns: bool = False,
) -> HttpResponse:
    exclude_columns = {"pk", "actions"}

    if use_selected_columns:
        exclude_columns |= {name for name, _ in table.available_columns}

    exporter = TableExport(
        export_format=TableExport.CSV,
        table=table,
        exclude_columns=exclude_columns,
    )
    return exporter.response(filename=filename)


def is_hex_string(s: str, min_octets: int, max_octets: int):
    if not re.match(constants.HEX_STRING_REGEX, s):
        return False

    octets = len(s.replace(":", "").replace("-", "")) / 2
    return octets >= min_octets and octets <= max_octets


def check_dhcp_enabled(instance: Server, version: Literal[6, 4]) -> HttpResponse | None:
    if (version == 6 and instance.dhcp6) or (version == 4 and instance.dhcp4):
        return None
    return redirect(instance.get_absolute_url())


class OptionalViewTab(ViewTab):
    def __init__(self, *args, is_enabled: Callable[[Any], bool], **kwargs) -> None:
        self.is_enabled = is_enabled
        super().__init__(*args, **kwargs)

    def render(self, instance):
        if self.is_enabled(instance):
            return super().render(instance)
        return None



import django_tables2 as tables
from django.urls import reverse
from django.utils.http import urlencode
from netbox.tables import BaseTable, BooleanColumn, NetBoxTable, ToggleColumn, columns

from netbox_kea.utilities import format_duration

from .models import Server

SUBNET_ACTIONS = """<span class="btn-group dropdown">
  <a class="btn btn-sm btn-secondary dropdown-toggle" href="#" type="button" data-bs-toggle="dropdown">
  <i class="mdi mdi-magnify"></i></a>
  <ul class="dropdown-menu">
    {% if record.pk %}
    <li>
      <a href="{% url "ipam:prefix" pk=record.pk %}" class="dropdown-item">
        <i class="mdi mdi-open-in-app" aria-hidden="true" title="View prefix"></i>
        View prefix
      </a>
    </li>
    {% endif %}
    {% if record.subnet %}
    <li>
      <a href="{% url "ipam:prefix_list" %}?prefix={{ record.subnet }}" class="dropdown-item">
        <i class="mdi mdi-magnify" aria-hidden="true" title="Search for prefix"></i>
        Search for prefix
      </a>
    </li>
    {% endif %}
  </ul>
</span>
"""  # noqa: E501


LEASE_ACTIONS = """<span class="btn-group dropdown">
    <a class="btn btn-sm btn-secondary dropdown-toggle" href="#" type="button" data-bs-toggle="dropdown">
    <i class="mdi mdi-magnify"></i></a>
    <ul class="dropdown-menu">
        {% if record.ip_address %}
        <li>
            <a href="{% url "ipam:ipaddress_list" %}?address={{ record.ip_address }}" class="dropdown-item">
                <i class="mdi mdi-magnify" aria-hidden="true" title="Search IPs"></i>
                Search IPs
            </a>
        </li>
        {% endif %}
        {% if record.hw_address %}
        <li>
            <a href="{% url "dcim:interface_list" %}?mac_address={{ record.hw_address }}" class="dropdown-item">
                <i class="mdi mdi-magnify" aria-hidden="true" title="Search interfaces"></i>
                Search interfaces
            </a>
        </li>
        <li>
            <a href="{% url "virtualization:vminterface_list" %}?mac_address={{ record.hw_address }}" class="dropdown-item">
                <i class="mdi mdi-magnify" aria-hidden="true" title="Search VM interfaces"></i>
                Search VM interfaces
            </a>
        </li>
        {% endif %}
        {% if record.hostname %}
        <li>
            <a href="{% url "dcim:device_list" %}?q={{ record.hostname }}" class="dropdown-item">
                <i class="mdi mdi-magnify" aria-hidden="true" title="Search devices"></i>
                Search devices
            </a>
        </li>
        <li>
            <a href="{% url "virtualization:virtualmachine_list" %}?q={{ record.hostname }}" class="dropdown-item">
                <i class="mdi mdi-magnify" aria-hidden="true" title="Search VMs"></i>
                Search VMs
            </a>
        </li>
        {% endif %}
    </ul>
</span>
"""  # noqa: E501


class DurationColumn(tables.Column):
    def render(self, value: int):
        """Value is in seconds."""
        return format_duration(value)


class ActionsColumn(tables.TemplateColumn):
    def __init__(self, template: str) -> None:
        super().__init__(
            template,
            attrs={"td": {"class": "text-end text-nowrap noprint"}},
            verbose_name="",
        )


class MonospaceColumn(tables.Column):
    def __init__(self, *args, additional_classes: list[str] | None = None, **kwargs):
        cls_str = "font-monospace"
        if additional_classes is not None:
            cls_str += " " + " ".join(additional_classes)
        super().__init__(*args, attrs={"td": {"class": cls_str}}, **kwargs)


class ServerTable(NetBoxTable):
    name = tables.Column(linkify=True)
    dhcp6 = BooleanColumn()
    dhcp4 = BooleanColumn()

    class Meta(NetBoxTable.Meta):
        model = Server
        fields = (
            "pk",
            "name",
            "server_url",
            "username",
            "password",
            "ssl_verify",
            "client_cert_path",
            "client_key_path",
            "ca_file_path",
            "dhcp6",
            "dhcp4",
        )
        default_columns = ("pk", "name", "server_url", "dhcp6", "dhcp4")


# we can't use NetBox table because it requires an actual model
class GenericTable(BaseTable):
    exempt_columns = ("actions", "pk")

    class Meta(BaseTable.Meta):
        empty_text = "No rows"
        fields: tuple[str, ...] = ()

    @property
    def objects_count(self):
        return len(self.data)


class SubnetTable(GenericTable):
    id = tables.Column(verbose_name="ID")
    subnet = tables.Column(
        linkify=lambda record, table: (
            (
                reverse(
                    f"plugins:netbox_kea:server_leases{record['dhcp_version']}",
                    args=[record["server_pk"]],
                )
                + "?"
                + urlencode({"by": "subnet", "q": record["subnet"]})
            )
            if record.get("subnet")
            else None
        ),
    )
    shared_network = tables.Column(verbose_name="Shared Network")
    actions = ActionsColumn(SUBNET_ACTIONS)

    class Meta(GenericTable.Meta):
        empty_text = "No subnets"
        fields = ("id", "subnet", "shared_network", "actions")
        default_columns = ("id", "subnet", "shared_network")


class BaseLeaseTable(GenericTable):
    # This column is for the select checkboxes.
    pk = ToggleColumn(verbose_name="IP Address", accessor="ip_address", visible=True)
    ip_address = tables.Column(verbose_name="IP Address")
    hostname = tables.Column(verbose_name="Hostname")
    subnet_id = tables.Column(verbose_name="Subnet ID")
    hw_address = MonospaceColumn(verbose_name="Hardware Address")
    valid_lft = DurationColumn(verbose_name="Valid Lifetime")
    cltt = columns.DateTimeColumn(verbose_name="Client Last Transaction Time")
    expires_at = columns.DateTimeColumn(verbose_name="Expires At")
    expires_in = DurationColumn(verbose_name="Expires In")
    actions = ActionsColumn(LEASE_ACTIONS)

    class Meta(GenericTable.Meta):
        empty_text = "No leases found."
        fields = (
            "ip_address",
            "hostname",
            "subnet_id",
            "hw_address",
            "valid_lft",
            "cltt",
            "expires_at",
            "expires_in",
            "actions",
        )
        default_columns = ("ip_address", "hostname")


class LeaseTable4(BaseLeaseTable):
    client_id = tables.Column(verbose_name="Client ID")

    class Meta(BaseLeaseTable.Meta):
        fields = ("client_id", *BaseLeaseTable.Meta.fields)


class LeaseTable6(BaseLeaseTable):
    type = tables.Column(verbose_name="Type", accessor="type")
    preferred_lft = DurationColumn(verbose_name="Preferred Lifetime")
    duid = MonospaceColumn(verbose_name="DUID", additional_classes=["text-break"])
    iaid = MonospaceColumn(verbose_name="IAID")

    class Meta(BaseLeaseTable.Meta):
        fields = ("type", "duid", "iaid", *BaseLeaseTable.Meta.fields)


class LeaseDeleteTable(GenericTable):
    ip_address = tables.Column(verbose_name="IP Address", accessor="ip")

    class Meta(NetBoxTable.Meta):
        empty_text = "No leases"
        fields = ("ip_address",)
        default_columns = ("ip_address",)
